/*
** Module   : ConsoleHelper.java
** Abstract : The class provides JNI calls to text console based C code. 
**
** Modifications Copyright (c) 2005-2010, Golden Code Development Corporation.
** The original LGPL license terms apply.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20130723 Created initial version
*/

package com.goldencode.p2j.ui.client.chui.driver.console;

import java.io.FileDescriptor;

/**
 * The ConsoleHelper class provides the interface to the "ncurses" library
 * functions. Exceptions and error messages are reported to the logfile
 * $HOME/charva.log.
 * <p>
 * The default colors are white foreground and black background, but
 * these defaults can be modified by changing the static variables
 * _defaultForeground and _defaultBackground.
 */
public class ConsoleHelper 
{
   private static ConsoleHelper instance;
   
   /**
    * This flag is true if the system property "console.color" has been set.
    */
   public static final boolean isColorEnabled =
      (System.getProperty("console.color") != null);

   /**
    * Constructor which can only be called internally. Implements the singleton
    * pattern.
    */
   private ConsoleHelper()
   {
   }

   /**
    * Initialize the native layer.
    */
   public static void initChui()
   {
      System.loadLibrary("p2j");
      initConsole(); // call native function to initialize console.
   }

   /**
    * Access the singleton instance of the toolkit.
    *
    * @return   The only toolkit instance.
    */
   public synchronized static ConsoleHelper getDefaultConsoleHelper()
   {
      if (instance == null) {
         instance = new ConsoleHelper();
      }
      return instance;
   }
   
   /**
    * This method is used for initializing the line-drawing
    * characters using CURSES/NCURSES.
    */
   public static native int getACSchar(int offset);

   /**
    * This method is used for initializing the CURSES/NCURSES video
    * attributes.
    */
   public static native int getAttribute(int offset);

   /**
    * This method is intended to be called by the constructor.
    */
   private static native void initConsole();

   /**
    * Perform a bulk output operation for the whole screen at once with
    * output masking when desired. Set the cursor afterwards.
    * <p>
    * The specified array safe for use with characters that are UNICODE
    * encoded which are also outside of the 7-bit ASCII range. Such
    * characters are passed through to the native layer directly and are
    * not converted to a natively encoded form.  This could be provided
    * but since one of the primary uses of this method is for drawing
    * characters (e.g. <code>ACS_HLINE</code>) which are encoded using the
    * full 32-bits, any attempt to convert these into a single byte native
    * encoding would lose data.
    * <p>
    * Because of the above limitation, DO NOT USE THIS FOR OUTPUT OF USER
    * DATA THAT COULD CONTAIN DATA OUTSIDE OF THE 7-bit ASCII RANGE.
    *
    * @param    chr
    *           The array of characters and attributes to output. The array
    *           is supposed to provide an item per every screen position.
    *           This item is either -1 if this position is masked or a code
    *           which goes directly to ncurses - a combination of character,
    *           attribute and color.
    * @param    cursorX
    *           final cursor position, column
    * @param    cursorY
    *           final cursor position, row
    */
   public native void addArrayNative(int chr[], int cursorX, int cursorY);

   /**
    * Add a string to the virtual terminal at the current cursor position.
    */
   public native void addStringNative(byte[] str,
                                      int    len,
                                      int    attrib,
                                      int    colorpair);

   /**
    * Ring the terminal's bell.
    */
   public native void beep();

   /**
    * Clears the screen (calls the curses clear() function).
    */
   public native void clear();

   /**
    * Gets the number of columns on the screen.
    *
    * @return   The number of columns (in characters).
    */
   public native int getScreenColumnsNative();

   /**
    * Gets the number of rows on the screen.
    *
    * @return   The number of rows (lines of characters).
    */
   public native int getScreenRowsNative();
   
   /**
    * Returns true if the terminal is capable of displaying colours.
    */
   public native boolean hasColors();

   /**
    * Set absolute cursor position
    * 
    * @return    <code>true</code> if cursor positioning succeeded.
    */
   public native boolean setCursorNative(int x, int y);

   /**
    * Synchronize the state of the physical terminal with the state of
    * the virtual terminal maintained by curses (i.e. provides an interface
    * to the curses refresh() function).
    */
   public native void syncNative();
   
   /**
    * Sets the terminal type used by NCURSES.
    *
    * @param    name
    *           The new terminal type (must not be <code>null</code>).
    */
   public static native void setTermType(String name);
   
   /**
    * Gets the terminal type reported by NCURSES.
    *
    * @return   The terminal type or an empty string on any failure.
    */
   public static native String getTermType();
   
   /**
    * Set the cursor visibility on or off. Note that can invisible cursor
    * can still be moved (with {@link #setCursorNative} and used normally. This
    * method only affects the visibility of the cursor.
    *
    * @param    on
    *           When <code>true</code> the cursor will become visible, when
    *           <code>false</code> the cursor will become invisible.
    *
    * @return   The previous state of the cursor (<code>true</code> means
    *           visible).
    */
   public native boolean setCursorStatus(boolean on)
   throws IllegalStateException;

   /**
    * This method is intended to be called by the keyboard-reading thread
    * only.
    */
   public native int readKey();

   /**
    * Get current X position of cursor.
    */
   public native int getx();

   /**
    * Get current Y position of cursor.
    */
   public native int gety();

   /**
    * Switch back to the NCURSES mode after the temporary suspension.
    */
   public native void resume();

   /**
    * Save the current state of the NCURSES mode and switch to TTY mode
    * temporarily.
    */
   public native void suspend();

   /**
    * Reset terminal settings. May be called only if suspend() was previously
    * called. 
    */
   public native void resetMode();
}
