/*
** Module   : terminal_linux.c
** Abstract : Linux implementation for character mode-based console functions.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 EVL 20130723 Created initial version to implement text console terminal JNI.
*/

#include "terminal.h"

// pointer to the structure of the interactive terminal
static SCREEN* terminal = NULL;

static int hascolors = 0;	      // set when we initialize
static int colors_started = 0;	// set when we call start_color().

#define ACS_MAX 18
static jint acs_result[ACS_MAX];

// terminal size at the moment of initialization of the library 
extern int base_rows;
extern int base_cols;

/* Local functions declaration*/
void throwException(JNIEnv* env, char* name, char* msg)
{
   jclass jcls = (*env)->FindClass(env, name);
   
   if (jcls != NULL)
   {
      (*env)->ThrowNew(env, jcls, msg);
   }
}

/* This function outputs the screenful of data at once using an array of codes
   where every character position is either a skip indicator (-1) or a fully 
   prepared code with character, attribute and color included.
   The final step is the cursor positioning.
*/
JNIEXPORT void JNICALL addArrayNative(JNIEnv *env, jobject jo, jintArray arr, jint cx, jint cy)
{
   int i, j, chr;
   jboolean isCopy = JNI_FALSE;

   jint* arrp = (*env)->GetIntArrayElements(env, arr, &isCopy);
   jint* arp  = arrp;

   if (arp == NULL)
   {
      return;
   }

   for (i = 0; i < base_rows; i++)
   {
      for (j = 0; j < base_cols; j++)
      {
         chr = *arp++;

         if (chr == -1)
         {
            continue;
         }

         move(i, j);
         addch(chr);
       }
   }

   move(cy, cx);
   refresh();

   (*env)->ReleaseIntArrayElements(env, arr, arrp, JNI_ABORT);
}

/*
 * Draw a string of text.
 */
JNIEXPORT void JNICALL addStringNative(JNIEnv *env, jobject jo,
                                       jbyteArray jstr, jint len, jint attrib, jint colorpair)
{
   int i;
   int attr = attrib;

   if (colors_started)
   {
      attr |= COLOR_PAIR(colorpair);
   }
   
   if (len <= 0)
   {
      return;
   }
   
   jbyte* chrs = (jbyte*) malloc(len * sizeof(jbyte));
   
   if (chrs == NULL)
   {
      throwException(env, "java/lang/OutOfMemoryError", "Cannot copy data.");
   }
   
   (*env)->GetByteArrayRegion(env, jstr, 0, len, chrs);
   
   for (i = 0; i < len; i++)
   {
      // remove the automatic sign extension (we want to treat this as
      // an unsigned 8-bit number)
      int extended = 0x000000FF & (int) chrs[i];
      addch(extended | attr);
   } 

   (*env)->DeleteLocalRef(env, jstr);
   free(chrs);
}

/*
 * Ring the terminal bell.
 */
JNIEXPORT void JNICALL beepNative(JNIEnv *env, jobject jo)
{
   beep();
}

JNIEXPORT void JNICALL clearNative(JNIEnv *env, jobject jo)
{
   clear();
}

JNIEXPORT jint JNICALL getACSchar(JNIEnv *env, jclass class, jint offset)
{
   if (offset >= 0 && offset < ACS_MAX)
   {
      return acs_result[offset];
   }
   else
   {
      return (jint) -1;
   }
}

#define ATTRIBUTE_MAX 9

JNIEXPORT jint JNICALL getAttribute(JNIEnv *env, jclass class, jint offset)
{
   static jint attribute_result[ATTRIBUTE_MAX] = {A_NORMAL, A_STANDOUT, A_UNDERLINE, A_REVERSE,
                                                  A_BLINK, A_DIM, A_BOLD, A_ALTCHARSET, A_INVIS};

   if (offset >= 0 && offset < ATTRIBUTE_MAX)
   {
      return attribute_result[offset];
   }
   else
   {
      return (jint) -1;
   }
}

// returns the termname() or "" on any failure
char* safe_termname()
{
   char* term = termname();
   return (term == NULL) ? "" : term;
}

// return the NCURSES terminal type
JNIEXPORT jstring JNICALL getTermType(JNIEnv *env, jclass  cls)
{
   return (*env)->NewStringUTF(env, safe_termname());
}

/* Get the x position of the cursor.
 */
JNIEXPORT jint JNICALL getx(JNIEnv *env, jobject jo)
{
   int x, y;

   getyx(stdscr, y, x);
   return x;
}

/* Get the y position of the cursor.
 */
JNIEXPORT jint JNICALL gety(JNIEnv *env, jobject jo)
{
   int x, y;

   getyx(stdscr, y, x);
   return y;
}

/* Returns true if the terminal is capable of displaying colors.
 */
JNIEXPORT jboolean JNICALL hasColors(JNIEnv *env, jobject jo)
{
   return has_colors();
}

JNIEXPORT void JNICALL initConsole(JNIEnv *env, jobject jo)
{
   acs_result[0]  = ACS_ULCORNER;
	acs_result[1]  = ACS_LLCORNER;
	acs_result[2]  = ACS_URCORNER;
	acs_result[3]  = ACS_LRCORNER;
	acs_result[4]  = ACS_LTEE;
	acs_result[5]  = ACS_RTEE;
	acs_result[6]  = ACS_BTEE;
	acs_result[7]  = ACS_TTEE;
	acs_result[8]  = ACS_HLINE;
	acs_result[9]  = ACS_VLINE;
	acs_result[10] = ACS_PLUS;
	acs_result[11] = ACS_S1;
	acs_result[12] = ACS_S9;
	acs_result[13] = ACS_DIAMOND;
	acs_result[14] = ACS_CKBOARD;
	acs_result[15] = ACS_DEGREE;
	acs_result[16] = ACS_PLMINUS;
	acs_result[17] = ACS_BULLET;
   
// EVL<<
// EVL***   char *strcap;
// EVL>>
   char *termname;
// EVL<<
// EVL***   int i;
// EVL>>

   termname = getenv("TERM");
    
   if (termname == NULL || *termname == '\0')
   {
      termname = "vt100";
   }
   
   // use this instead of initscr because there is no way to query the
   // current terminal and when you use multiple terms, newterm overwrites
   // the current terminal... so you can never get it if you don't use
   // newterm directly
   terminal = newterm(termname, stdout, stdin);
    
   if (terminal == NULL)
   {
      char* spec   = "Interactive terminal of type %s cannot be initialized.";
      char* errmsg = malloc(strlen(spec) + strlen(termname) + 1); 
      sprintf(errmsg, spec, termname);
      throwException(env, "java/lang/IllegalStateException", errmsg);
      return;
   }
    
   keypad(stdscr, TRUE);      // enable keyboard mapping
   timeout(500);              // wait up to 500msec for input
   noecho();                  // don't echo input
   raw();
   nonl();
   auto_getch_refresh(FALSE); // disable sync during getch()
   def_prog_mode();           // save our state (after our mode changes)

   hascolors = has_colors();

   getmaxyx(stdscr, base_rows, base_cols);

   atexit((void (*)(void))endwin);
}

// Returns -1 if there was no character to read.
// Otherwise returns the character that was read.
JNIEXPORT jint JNICALL readKey(JNIEnv *env, jobject jo)
{
   return (jint) getch();
}

JNIEXPORT void JNICALL resetMode(JNIEnv *env, jobject jo)
{
   reset_prog_mode();
}

JNIEXPORT void JNICALL resume(JNIEnv *env, jobject jo)
{
   reset_prog_mode();          /* Return to the previous tty mode*/
                               /* stored by def_prog_mode()      */
   doupdate();                 /* Do refresh() to restore the    */
                               /* Screen contents                */
}

JNIEXPORT jboolean JNICALL setCursorNative(JNIEnv *env, jobject jo, jint x, jint y)
{
   int localX, localY;
   move(y, x);

   // Set current cursor position
   getyx(stdscr, localY, localX);
    
   return localY == y && localX == x;
}

JNIEXPORT jboolean JNICALL setCursorStatus(JNIEnv *env, jobject jo, jboolean on)
{
   int rc = ERR;
   jboolean ret;
   
   if( on )
      rc = curs_set( 1 );
   else
      rc = curs_set( 0 );
   
   if ( rc == 1 || rc == 2 )
      ret = true;
   else if ( rc == 0 )
      ret = false;
   else
   {
      char* spec;
      char* termname = getenv("TERM");
      if ( rc == ERR )
         return true;
      else
         spec = "Unexpected return from curs_set() for terminal of type %s (%d).";
      char* errmsg = malloc(strlen(spec) + strlen(termname) + 32); 
      sprintf(errmsg, spec, termname, rc);
      throwException(env, "java/lang/IllegalStateException", errmsg);
   }
   
   return ret;
}

// simple reset of terminal type
void restartTerminal(JNIEnv *env, char* termname)
{
   int   err  = 0;
   int   rc   = restartterm(termname, fileno(stdout), &err); 
   
   if (rc == ERR)
   {
      char* spec   = "Failure resetting terminal of type %s (error %i).";
      char* errmsg = malloc(strlen(spec) + strlen(termname) + 15); 
      sprintf(errmsg, spec, termname, err);
      throwException(env, "java/lang/IllegalStateException", errmsg);
   }
}

// set the NCURSES terminal type
JNIEXPORT void JNICALL setTermType(JNIEnv *env, jclass  cls, jstring name)
{
   char* termname = (char*) (*env)->GetStringUTFChars(env, name, NULL);
   
   if (termname == NULL)
   {
      // out of memory!
      exit(-1);
   }
   
   restartTerminal(env, strdup(termname));
   
   (*env)->ReleaseStringUTFChars(env, name, termname);
}

JNIEXPORT void JNICALL suspend(JNIEnv *env, jobject jo)
{
   def_prog_mode();            /* Save the tty modes             */
   endwin();                   /* End curses mode temporarily    */
}

JNIEXPORT void JNICALL syncNative(JNIEnv *env, jobject jo)
{
   refresh();    
}
